/*
  'Acqua' kwin client

  Copyright (C) 2001 Rik Hemsley (rikkus) <rik@kde.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
*/

#include <unistd.h> // for usleep
#include <math.h>

#include <qlayout.h>
#include <qpainter.h>

#include <netwm.h>
#include <kwin/workspace.h>
#undef Bool

#include <kconfig.h>
#include <kstddirs.h>

#include <kapp.h>
#include <kwin/options.h>

#include "Acqua.h"
#include "AcquaButton.h"
#include "AcquaButtonHelp.h"
#include "AcquaButtonIconify.h"
#include "AcquaButtonClose.h"
#include "AcquaButtonMenu.h"
#include "AcquaButtonMaximize.h"

using namespace KWinInternal;

extern "C"
{
  Client * allocate(Workspace * ws, WId w, int tool)
  {
    return new Acqua(ws, w, tool != 0);
  }

  void init()
  {
    // Empty.
  }

  void reset()
  {
    Workspace::self()->slotResetAllClientsDelayed();
    // Empty.
  }

  void deinit()
  {
    // Empty.
  }
}

const uint sideMargin    = 2;
const uint topSideMargin = 4;
const uint topMargin     = 4;
const uint bottomMargin  = 4;
const uint textVMargin   = 2;


Acqua::Acqua(Workspace * ws, WId w, bool tool, QWidget * parent, const char * name)
  : Client        (ws, w, parent, name, WResizeNoErase),
    tool_         (tool),
    mainLayout_   (0),
    titleSpacer_  (0)
{
  setBackgroundMode(NoBackground);

  _resetLayout();

  leftButtonList_   .setAutoDelete(true);
  rightButtonList_  .setAutoDelete(true);

  connect(options, SIGNAL(resetClients()), this, SLOT(slotReset()));
}

Acqua::~Acqua()
{
  // Empty.
}

  void
Acqua::resizeEvent(QResizeEvent * e)
{
  Client::resizeEvent(e);
  doShape();
  repaint();
}

  void
Acqua::captionChange(const QString &)
{
  repaint();
}

  void
Acqua::paintEvent(QPaintEvent * pe)
{
  QRect titleRect(1, 1, width()-2, titleSpacer_->geometry().height()+topMargin*2-1);

  QPainter p(this);

  p.setPen(Qt::black);
  p.setBrush(colorGroup().background());

  p.drawRect(rect());

  p.setClipRegion(pe->region() - titleRect);

  p.setClipRegion(pe->region());

  QListIterator<AcquaButton> it(rightButtonList_);
  QRect brect(1, it.current()->geometry().top(), width()-2, it.current()->geometry().height());

  /* The folowing lines are a bit strange, because i have to match the
   * background of the buttons. If anybody can provid some transparent
   * backgound button image, it can be quicker and more graceful.
   */
  QColor cols[16];
  int xx=width()-2, hh=titleSpacer_->geometry().height()+topMargin*2-2;
  if(isActive()){
    cols[0].setRgb(241, 241, 241);
    cols[1].setRgb(255, 255, 255);
    cols[2].setRgb(241, 241, 241);
    cols[3].setRgb(233, 233, 233);
    
    for(int i=brect.top()-1, ci=0; i>0; i--, ci++)
    {
      if(ci>=4)ci=0;
      p.setPen(cols[ci]);
      p.drawLine(1, i, xx, i);
    }
    
    cols[0].setRgb(0xE9, 0xE9, 0xE9);
    cols[1].setRgb(0xF1, 0xF1, 0xF1);
    cols[2].setRgb(0xFF, 0xFF, 0xFF);
    cols[3].setRgb(0xF1, 0xF1, 0xF1);
    cols[4].setRgb(0xE9, 0xE9, 0xE9);
    cols[5].setRgb(0xEE, 0xEE, 0xEE);
    cols[6].setRgb(0xFD, 0xFD, 0xFD);
    cols[7].setRgb(0xEF, 0xEF, 0xEF);
    cols[8].setRgb(0xE6, 0xE6, 0xE6);
    cols[9].setRgb(0xEA, 0xEA, 0xEA);
    cols[10].setRgb(0xF7, 0xF7, 0xF7);
    cols[11].setRgb(0xE8, 0xE8, 0xE8);
    cols[12].setRgb(0xDE, 0xDE, 0xDE);
    cols[13].setRgb(0xE0, 0xE0, 0xE0);
    cols[14].setRgb(0xEA, 0xEA, 0xEA);
    cols[15].setRgb(0xD9, 0xD9, 0xD9);
    for(int i=brect.top(), ci=0; i<=brect.bottom(); i++, ci++)
    {      
      p.setPen(cols[ci]);
      p.drawLine(1, i, xx, i);
    }

    cols[0].setRgb(214, 214, 214);
    cols[1].setRgb(217, 217, 217);
    cols[2].setRgb(234, 234, 234);
    cols[3].setRgb(217, 217, 217);

    for(int i=brect.bottom()+1, ci=0; i<hh; i++, ci++){
      if(ci>=4)ci=0;
      p.setPen(cols[ci]);
      p.drawLine(1, i, xx, i);
    }
    
    p.setPen(QColor(200, 200, 200));
    p.drawLine(1, hh, xx, hh);
    
    p.setPen(QColor(196, 196, 196));
    p.drawLine(1, hh+1, xx, hh+1);
  }
  else {
    cols[0].setRgb(150, 169, 203);
    cols[1].setRgb(157, 179, 208);
    cols[2].setRgb(150, 169, 203);
    cols[3].setRgb(151, 167, 200);
    
    for(int i=brect.top()-1, ci=0; i>0; i--, ci++)
    {
      if(ci>=4)ci=0;
      p.setPen(cols[ci]);
      p.drawLine(1, i, xx, i);
    }
   
    cols[0].setRgb(0x97, 0xA7, 0xC8);
    cols[1].setRgb(0x96, 0xA9, 0xCB);
    cols[2].setRgb(0x9D, 0xB3, 0xD0);
    cols[3].setRgb(0x93, 0xAC, 0xC2);
    cols[4].setRgb(0x92, 0xA6, 0xC3);
    cols[5].setRgb(0x95, 0xAA, 0xC6);
    cols[6].setRgb(0x9E, 0xB2, 0xCF);
    cols[7].setRgb(0x93, 0xA7, 0xC4);
    cols[8].setRgb(0x93, 0xA7, 0xC4);
    cols[9].setRgb(0x93, 0xA7, 0xC4);
    cols[10].setRgb(0x9A, 0xAF, 0xCB);
    cols[11].setRgb(0x91, 0xA5, 0xC2);
    cols[12].setRgb(0x8C, 0xA5, 0xC3);
    cols[13].setRgb(0x91, 0xA6, 0xC0);
    cols[14].setRgb(0x9A, 0xAF, 0xCB);
    cols[15].setRgb(0x8D, 0xA0, 0xC0);

    for(int i=brect.top(), ci=0; i<=brect.bottom(); i++, ci++)
    {
      p.setPen(cols[ci]);
      p.drawLine(1, i, xx, i);
    }

    cols[0].setRgb(137, 161, 194);
    cols[1].setRgb(137, 163, 192);
    cols[2].setRgb(154, 175, 203);
    cols[3].setRgb(141, 160, 192);

    for(int i=brect.bottom()+1, ci=0; i<hh; i++, ci++){
      if(ci>=4)ci=0;
      p.setPen(cols[ci]);
      p.drawLine(1, i, xx, i);
    }
    
    p.setPen(QColor(132, 160, 190));
    p.drawLine(1, hh, xx, hh);
  }
  /* end of the strange code. */
  
  p.setPen(QColor(96, 96, 96));
  p.drawLine(1, hh+1, xx, hh+1);

  p.setPen(Qt::black);
  if (shape_)
  {
    int r(width());
    int b(height());

    // Draw edge of top-left corner inside the area removed by the mask.

    p.drawPoint(3, 1);
    p.drawPoint(4, 1);
    p.drawPoint(2, 2);
    p.drawPoint(1, 3);
    p.drawPoint(1, 4);

    // Draw edge of top-right corner inside the area removed by the mask.

    p.drawPoint(r - 5, 1);
    p.drawPoint(r - 4, 1);
    p.drawPoint(r - 3, 2);
    p.drawPoint(r - 2, 3);
    p.drawPoint(r - 2, 4);
  }

  p.setFont(options->font(isActive(), tool_));

  p.setPen(options->color(Options::Font, isActive()));

  p.drawText(titleSpacer_->geometry(), AlignCenter, caption());
}

  void
Acqua::doShape()
{
  if (!shape_)
    return;
  
  QRegion mask(0, 0, width(), height());

  int r(width());
  int b(height());

  // Remove top-left corner.

  mask -= QRegion(0, 0, 5, 1);
  mask -= QRegion(0, 1, 3, 1);
  mask -= QRegion(0, 2, 2, 1);
  mask -= QRegion(0, 3, 1, 2);

  // Remove top-right corner.

  mask -= QRegion(r - 5, 0, 5, 1);
  mask -= QRegion(r - 3, 1, 3, 1);
  mask -= QRegion(r - 2, 2, 2, 1);
  mask -= QRegion(r - 1, 3, 1, 2);

  setMask(mask);
}

  void
Acqua::showEvent(QShowEvent *)
{
  doShape();
  repaint();
}

  void
Acqua::windowWrapperShowEvent(QShowEvent *)
{
  doShape();
  repaint();
}

  void
Acqua::mouseDoubleClickEvent(QMouseEvent * e)
{
  if (titleSpacer_->geometry().contains(e->pos()))
  {
    workspace()
      ->performWindowOperation(this, options->operationTitlebarDblClick());
  }
}

  void
Acqua::stickyChange(bool b)
{
}

  void
Acqua::maximizeChange(bool b)
{
  emit(maxChange(b));
}

  void
Acqua::activeChange(bool)
{
  for (QListIterator<AcquaButton> it(leftButtonList_); it.current(); ++it)
  {
    it.current()->repaint();
  }

  for (QListIterator<AcquaButton> it(rightButtonList_); it.current(); ++it)
  {
    it.current()->repaint();
  }

  repaint();
}

  Client::MousePosition
Acqua::mousePosition(const QPoint & p) const
{
  int x = p.x();
  int y = p.y();

  if (y < titleSpacer_->geometry().height()+8)
  {
      return Client::Center;
  }
  else if (y < height() - 20)
  {
    if (x < 4)
      return Client::Left;
    else
      return Client::Right;
  }
  else
  {
    if (x < 20)
      return Client::BottomLeft;
    else
      if (x > width() - 20)
        return Client::BottomRight;
      else
        return Client::Bottom;
  }

  return Client::mousePosition(p);
}

  void
Acqua::slotReset()
{
  _resetLayout();
  repaint();
}

  void
Acqua::slotMaximize(int button)
{
  switch (button)
  {
    case MidButton:
      maximize(MaximizeVertical);
      break;

    case RightButton:
      maximize(MaximizeHorizontal);
      break;

    case LeftButton:
    default:
      maximize(MaximizeFull);
  }
}

void 
Acqua::slotPopupMenu(QPoint pos)
{
  workspace()->clientPopup(this)->popup(pos);
}

AcquaButton *
Acqua::_createButton(const QString & s, QWidget * parent)
{
  AcquaButton * b = 0;

  if (("Help" == s) && providesContextHelp())
  {
    b = new AcquaButtonHelp(parent);
    connect(b, SIGNAL(help()), this, SLOT(contextHelp()));
  }
  else if ("Menu" == s)
  {
    b = new AcquaButtonMenu(parent);
    connect(b, SIGNAL(popupMenu(QPoint)), this, SLOT(slotPopupMenu(QPoint)));
  }
  else if ("Iconify" == s && isMinimizable())
  {
    b = new AcquaButtonIconify(parent);
    connect(b, SIGNAL(iconify()), this, SLOT(iconify()));
  }

  else if ("Maximize" == s && isMaximizable())
  {
    b = new AcquaButtonMaximize(isMaximized(), parent);
    connect(b, SIGNAL(maximize(int)), this, SLOT(slotMaximize(int)));
    connect(this, SIGNAL(maxChange(bool)), b, SLOT(slotMaximizeChange(bool)));
  }

  else if ("Close" == s)
  {
    b = new AcquaButtonClose(parent);
    connect(b, SIGNAL(closeWindow()), this, SLOT(closeWindow()));
  }

  return b;
}

  void
Acqua::_createButtons()
{
  leftButtonList_   .clear();
  rightButtonList_  .clear();

  QString buttons = options->titleButtonsLeft() + "|" + options->titleButtonsRight();
  QList<AcquaButton> *buttonList = &leftButtonList_;

  for (unsigned int i = 0; i < buttons.length(); ++i)
  {
    AcquaButton * tb = 0;
    switch (buttons[i].latin1())
    {
      case 'M': // Menu
        tb = _createButton("Menu", this);
        break;

      case 'H': // Help
//        tb = _createButton("Help", this);
        break;

      case 'I': // Minimize
        tb = _createButton("Iconify", this);
        break;

      case 'A': // Maximize
        tb = _createButton("Maximize", this);
        break;

      case 'X': // Close
        tb = _createButton("Close", this);
        break;

      case '|':
        buttonList = &rightButtonList_;
        break;
    }
    if (0 != tb)
      buttonList->append(tb);
  }
}

  void
Acqua::_resetLayout()
{
  KConfig c(locate("config", "kwinwebrc"));
  c.setGroup("General");
  shape_ = c.readBoolEntry("Shape", true);

  //  ____________________________________
  // |  |                              |  |
  // |Xo|     titleSpacer              |v^| <--- topLayout
  // |__|______________________________|__|
  // | |                                | |
  // | |                                | |
  // | |     windowWrapper              | |
  // | |                                | | <--- midLayout
  // | |                                | |
  // | |                                | |
  // | |________________________________| |
  // |____________________________________|

  QFontMetrics fm(options->font(isActive(), tool_));

  uint titleHeight = QMAX(14, fm.height());

  if (0 != titleHeight % 2)
    titleHeight += 1;

  delete mainLayout_;

  mainLayout_  = new QVBoxLayout(this, 0, 0);

  titleSpacer_ =
    new QSpacerItem
    (
     0,
     titleHeight,
     QSizePolicy::Expanding,
     QSizePolicy::Fixed
    );

  QHBoxLayout * topLayout   = new QHBoxLayout(mainLayout_, 0, 0);
  topLayout->setMargin(topMargin);

  _createButtons();

  // Add left-side buttons.

  topLayout->addSpacing(topSideMargin);
  
  for (QListIterator<AcquaButton> it(leftButtonList_); it.current(); ++it)
  {
    topLayout->addWidget(it.current(), Qt::AlignVCenter);
    topLayout->setStretchFactor(it.current(), 0);
  //  it.current()->setFixedSize(titleHeight, titleHeight);
  }

  topLayout->addItem(titleSpacer_);

  // Add right-side buttons.

  for (QListIterator<AcquaButton> it(rightButtonList_); it.current(); ++it)
  {
    topLayout->addWidget(it.current(), Qt::AlignVCenter);
  //  it.current()->setFixedSize(titleHeight, titleHeight);
  }
 
  topLayout->addSpacing(topSideMargin);
  // -------------------------------------------------------------------

  QHBoxLayout * midLayout   = new QHBoxLayout(mainLayout_, 0, 0);
 
  midLayout->addSpacing(sideMargin);
  midLayout->addWidget(windowWrapper());
  midLayout->addSpacing(sideMargin);

  // -------------------------------------------------------------------

  mainLayout_->addSpacing(bottomMargin);

  // Make sure that topLayout doesn't stretch - midLayout should take
  // all spare space.

  mainLayout_->setStretchFactor(topLayout, 0);
  mainLayout_->setStretchFactor(midLayout, 1);
}

  void
Acqua::animateIconifyOrDeiconify(bool /* iconify */)
{
  // Nice and simple ;)
}

#include "Acqua.moc"
// vim:ts=2:sw=2:tw=78:set et:
